#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
#include "True_Bmp.h"
#include "256_Bmp.h"

#define MODE_ABS 0
#define MODE_RLE 1

static int Width, Height;
static unsigned __int32 palette[256];
static unsigned __int32 *bytes;

int	Read256ColorBMP_RGB(BITMAPINFOHEADER *bmih, FILE *BMPfp)
{
    unsigned short BuffLength;
    unsigned short Color_Cnt;
	unsigned char *lpbBuf;
    RGBQUAD	palette[256];
    int	i, j;

	Width = bmih->biWidth;
	Height = bmih->biHeight;
	BuffLength = ((Width + 3) >> 2) << 2;
	Color_Cnt = 256;
	// do some calculations

    if((lpbBuf = (BYTE *)malloc(BuffLength * sizeof(BYTE))) == NULL)
		return 0;
	// alloc memory for scanline

	if((bytes = (unsigned __int32*)malloc(Width * Height * 4)) == NULL)
		return 0;
	// alloc memory for bitmap

    fread(palette, sizeof(RGBQUAD), Color_Cnt, BMPfp);
	// read palette

    for(i = Height - 1; i >= 0; i --) { 
        if(fread(lpbBuf, 1, BuffLength, BMPfp) != BuffLength) {
	        free(lpbBuf);
	        return 0;
        }
        for(j = 0; j < Width; j ++) {
	        bytes[j + Width * i] = (palette[lpbBuf[j]].rgbBlue << 16) +
				(palette[lpbBuf[j]].rgbGreen << 8) + palette[lpbBuf[j]].rgbRed;
		}
	}
	// read bitmap

	free((void*)lpbBuf);

    return 1;
}

int	Read256ColorBMP_RLE8(BITMAPINFOHEADER *bmih, FILE *BMPfp)
{
	unsigned short BuffLength;
    unsigned short Color_Cnt;
    unsigned char DummyCount;
	unsigned char *lpbBuf;
    RGBQUAD	palette[256];
    unsigned char Count;
    unsigned char bBuf;
    int	ReadCode, Mode;
    int	i, j;

	Width = bmih->biWidth;
	Height = bmih->biHeight;
	BuffLength = ((Width + 3) >> 2) << 2;
	Color_Cnt = 256;
	// do some calculations

	if((bytes = (unsigned __int32*) malloc(Width * Height * 4)) == NULL)
		return 0;
	// alloc memory for bitmap

    if((lpbBuf = (unsigned char*) malloc(BuffLength)) == NULL)
		return 0;
	// alloc memory for scanline

    fread(palette, sizeof(RGBQUAD), Color_Cnt, BMPfp);
	// read palette

    ReadCode = 1;
    for(i = Height - 1; i >= 0;) {
        if(ReadCode == 1) {
	        bBuf = fgetc(BMPfp);
	        ReadCode = 0;
	        if(bBuf == 0) {
	            bBuf = fgetc(BMPfp);
				switch(bBuf) {
				case 0x00:
					j = Width;
					ReadCode = 1;
					break;
				case 0x01:
					j = Width;
					i = 0;
					break;
				case 0x02:
					ReadCode = 1;
					bBuf = fgetc(BMPfp);
					j += bBuf;
					bBuf = fgetc(BMPfp);
					i -= bBuf;
					if(j >= Width) {
						j -= Width;
						i --;
					}
					break;
				default:
					Mode = MODE_ABS;
					Count = bBuf;
					break;
				}
			} else {
				Mode = MODE_RLE;
				Count = bBuf;
			}
		} else {
			if(Mode == MODE_ABS) {
				DummyCount = Count & 0x01;
				while(Count -- > 0) {
					bBuf = fgetc(BMPfp);
					if(j == Width) {
						j = 0;
						i --;
					}
					bytes[j + Width * i] = (palette[bBuf].rgbBlue << 16) +
						(palette[bBuf].rgbGreen << 8) + palette[bBuf].rgbRed;
					j ++;
				}
				if(DummyCount != 0)
					bBuf = fgetc(BMPfp);
			} else {
				bBuf = fgetc(BMPfp);
				while(Count -- > 0) {
					if(j == Width) {
						j = 0;
						i --;
					}
					bytes[j + Width * i] = (palette[bBuf].rgbBlue << 16) +
						(palette[bBuf].rgbGreen << 8) + palette[bBuf].rgbRed;
					j ++;
				}
			}
			ReadCode = 1;
		}
	}
	// read bitmap

	return 1;
}

// nacte 256barevnou bitmapu s / bez rle komprese.
BMP	*Load_256_ColorBMP(char *Cesta)
{
	BITMAPINFOHEADER bmih;
	BITMAPFILEHEADER bmfh;
	FILE *BMPfp;
	BMP *tmp;

	if((BMPfp = fopen(Cesta, "rb")) == NULL)
		return NULL;
	// open file

	fread(&bmfh, sizeof(bmfh), 1, BMPfp);
	fread(&bmih, sizeof(bmih), 1, BMPfp);
	// read headers

	if(bmfh.bfType != ('B' | ('M' << 8))) {
		fclose(BMPfp);
		return NULL;
	}
	// check image is bitmap

	if(bmih.biBitCount == 8 && bmih.biCompression == BI_RGB) {
		if(Read256ColorBMP_RGB(&bmih, BMPfp) == 0) {
			fclose(BMPfp);
			return NULL;
		}
	} else if(bmih.biBitCount == 8 && bmih.biCompression == BI_RLE8) {
		if(Read256ColorBMP_RLE8(&bmih, BMPfp) == 0) {
			fclose(BMPfp);
			return NULL;
		}
	} else
		return NULL;
	// read bitmap itself

	fclose(BMPfp);
	// close file

	if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
		return NULL;
	tmp->width = Width;
	tmp->height = Height;
	tmp->bytes = bytes;
	// create bitmap structure

	return tmp;
}
