#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
#include "True_Bmp.h"
#include "256_Pcx.h"

#define	PCX_PLANES	4

typedef struct {
    unsigned char  Id;
    unsigned char  Version;
    unsigned char  Format;
    unsigned char  BitsPixelPlane;
    unsigned short Xmin;
    unsigned short Ymin;
    unsigned short Xmax;
    unsigned short Ymax;
    unsigned short Hdpi;
    unsigned short Vdpi;
    unsigned char  EgaPalette[48];
    unsigned char  Reserved;
    unsigned char  NumberOfPlanes;
    unsigned short BytesLinePlane;
    unsigned short PaletteInfo;
    unsigned short HScreenSize;
    unsigned short VScreenSize;
    unsigned char  Filler[54];
} PCXHEADER;

typedef struct {
    unsigned char VgaPalette[768];
} PCXVGAPALETTE;

void ReadPcxHeader(PCXHEADER *PcxHead, FILE *FpPcx)
{
	// reading header
	fread(&PcxHead->Id, 1, 1, FpPcx);
	fread(&PcxHead->Version, 1, 1, FpPcx);
	fread(&PcxHead->Format, 1, 1, FpPcx);
	fread(&PcxHead->BitsPixelPlane, 1, 1, FpPcx);
	fread(&PcxHead->Xmin, 2, 1, FpPcx);
	fread(&PcxHead->Ymin, 2, 1, FpPcx);
	fread(&PcxHead->Xmax, 2, 1, FpPcx);
	fread(&PcxHead->Ymax, 2, 1, FpPcx);
	fread(&PcxHead->Hdpi, 2, 1, FpPcx);
	fread(&PcxHead->Vdpi, 2, 1, FpPcx);

    // Read the EGA Palette
    fread(PcxHead->EgaPalette, 1, 48, FpPcx);

	fread(&PcxHead->Reserved, 1, 1, FpPcx);
	fread(&PcxHead->NumberOfPlanes, 1, 1, FpPcx);
	fread(&PcxHead->BytesLinePlane, 2, 1, FpPcx);
	fread(&PcxHead->PaletteInfo, 2, 1, FpPcx);
	fread(&PcxHead->HScreenSize, 2, 1, FpPcx);
	fread(&PcxHead->VScreenSize, 2, 1, FpPcx);
    // Read the reserved area at the end of the header
	fread(PcxHead->Filler, 1, 54, FpPcx);
}

short PcxDecodeScanLine(unsigned char *DecodedBuffer, int BufferSize, FILE *FpPcx)
{
	static char rcount = 0;
    static char rvalue = 0;
	short total;
    int index;
    char byte;

	total = 0;
	index = 0;
	//
    do {
        for(total += rcount; rcount && index < BufferSize; rcount -- , index ++)
            DecodedBuffer[index] = rvalue;
		// decoding scanline

        if(rcount) {
            total -= rcount;
            return(total);
        } // return count of decoded pixels

		fread(&byte, 1, 1, FpPcx);
        if((byte & 0xC0) == 0xC0) {
            rcount = byte & 0x3F;
            fread(&rvalue, 1, 1, FpPcx);
        } else {
            rcount = 1;
            rvalue = byte;
        }
		// get and decode next pixel
    } while(index < BufferSize);

	// return count of decoded pixels
    return(total);
}

BMP *Load_256_ColorPCX(char *filename)
{
	unsigned char *decodedBuffer;
    unsigned char *imageBuffer;
	unsigned short numOfColors;
	PCXVGAPALETTE PcxVgaPal;
    unsigned short xSize;
    unsigned short ySize;
    short scanLineLength;
    PCXHEADER pcxHead;
    FILE *fpPcxIn;
    char int8[1];
    short ret;
	BMP *tmp;
	int i, n;

    if((fpPcxIn = fopen(filename, "rb")) == NULL)
		return NULL;
    // Open the input PCX image file

    ReadPcxHeader(&pcxHead, fpPcxIn);
    // Read the PCX image file header information

    if(pcxHead.Id != 0x0A)
        return NULL;
    // Check the identification byte value

    xSize = pcxHead.Xmax - pcxHead.Xmin + 1;
    ySize = pcxHead.Ymax - pcxHead.Ymin + 1;
    // Calculate size of image in pixels and scan lines

    numOfColors = (1 << (pcxHead.BitsPixelPlane * pcxHead.NumberOfPlanes));
    scanLineLength = pcxHead.NumberOfPlanes * pcxHead.BytesLinePlane;
    // Do some more calculations

    // Create an uncompressed scanline buffer
    if((decodedBuffer = (unsigned char *) malloc(scanLineLength)) == NULL)
        return NULL;
	if((imageBuffer = (unsigned char *) malloc(scanLineLength * ySize)) == NULL)
        return NULL;

    for(i = 0; i < ySize; i ++) {
        // Read scan line and decode it
        if((ret = PcxDecodeScanLine(decodedBuffer, scanLineLength, fpPcxIn)) != scanLineLength)
            printf("Error in decoded line length (was %d should be %d)\n", ret, scanLineLength);
		memcpy(&imageBuffer[i * xSize], decodedBuffer, scanLineLength);
		// copy data to image buffer
    }

    if(pcxHead.Version == 5) {
        fseek(fpPcxIn, -769L, SEEK_END);
		fread(int8, 1, 1, fpPcxIn);
        if(int8[0] == 0x0C) {
            // Read the VGA palette
            fread(&PcxVgaPal.VgaPalette, 1, 768, fpPcxIn);
			fclose(fpPcxIn);

			if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
				return NULL;
			if((tmp->bytes = (unsigned __int32*) malloc(4 * xSize * ySize)) == NULL)
				return NULL;
			// lock memotry for bitmap

			tmp->width = xSize;
			tmp->height = ySize;
			for(i = 0; i < xSize * ySize; i ++) {
				n = imageBuffer[i];
				tmp->bytes[i] = PcxVgaPal.VgaPalette[n * 3] +
					(PcxVgaPal.VgaPalette[n * 3 + 1] << 8) +
					(PcxVgaPal.VgaPalette[n * 3 + 2] << 16);
			}
			// creating truecolor bitmap

			free((void*)imageBuffer);
			free((void*)decodedBuffer);
			// freeing up memory
			
			return tmp;
        }
    }
    fclose(fpPcxIn);

	return NULL;
}
