//
//
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "True_Bmp.h"
#include "Gif.h"

#define NO_CODE -1

// GIF Header
typedef struct {
   char	signature[3];
   char version[3];
   unsigned short screenwidth;
   unsigned short screenheight;
   unsigned char  flags;
   unsigned char  background;
   unsigned char  aspect;
} GIFHEADER;

// GIF 89a Image Block descriptor
typedef struct {
   unsigned short left;
   unsigned short top;
   unsigned short width;
   unsigned short height;
   unsigned char  flags;
} IMAGEBLOCK;

// GIF 89a Graphic Control Extension Block
typedef struct {
    unsigned char  BlockSize;
    unsigned char  PackedField;
    unsigned short DelayTime;
    unsigned char  ColorIndex;
    unsigned char  Terminator;
} GIFGRAPHICCONTROL;

// GIF 89a Plain Text Extension Block
typedef struct {
    unsigned char  BlockSize;
    unsigned short TextGridLeft;
    unsigned short TextGridTop;
    unsigned short TextGridWidth;
    unsigned short TextGridHeight;
    unsigned char  CellWidth;
    unsigned char  CellHeight;
    unsigned char  TextFgColorIndex;
    unsigned char  TextBgColorIndex;
    unsigned char *PlainTextData;
    unsigned char  Terminator;
} GIFPLAINTEXT;

// GIF 89a Application Extension Block
typedef struct {
    unsigned char  BlockSize;
    unsigned char  Identifier[8];
    unsigned char  AuthentCode[3];
    unsigned char *ApplicationData;
    unsigned char  Terminator;
} GIFAPPLICATION;

// GIF 89a Comment Extension Block
typedef struct {
    unsigned char *CommentData;
    unsigned char  Terminator;
} GIFCOMMENT;

static int Width, Height;
static unsigned __int32 *bytes;
static unsigned char palette[768];

static int  wordmasktable[] = { 
	0x0000, 0x0001, 0x0003, 0x0007,
	0x000f, 0x001f, 0x003f, 0x007f,
	0x00ff, 0x01ff, 0x03ff, 0x07ff,
	0x0fff, 0x1fff, 0x3fff, 0x7fff };
static int inctable[] = {8, 8, 4, 2, 0};
static int startable[] = {0, 4, 2, 1, 0};

int UnpackImage(FILE *fp, int bits, int width, int height, int flags)
{
	unsigned char *firstcodestack;
	unsigned char *lastcodestack;
	unsigned char *linebuffer;
	unsigned char b[255];
	int *codestack;
    unsigned char *p;
    unsigned char *q;
	unsigned char *u;
	int currentcode;
	int codesize2;
	int blocksize;
	int codesize;
	int nextcode;
	int thiscode;
	int oldtoken;
	int line = 0;
	int byte = 0;
	int pass = 0;
	int bitsleft;
	int oldcode;
	int bits2;
	int xi;

    q = b;
	p = b;
    bitsleft = 8;
	//
    if(bits < 2 || bits > 8)
 	   return 0;
	//
    bits2 = 1 << bits;
    nextcode = bits2 + 2;
	codesize2 = 1 << (codesize = bits + 1);
	oldcode = NO_CODE;
	oldtoken = NO_CODE;
	//
	firstcodestack = (unsigned char*)malloc(4096);
	lastcodestack = (unsigned char*)malloc(4096);
	codestack = (int*)malloc(4096 * sizeof(int));
	if((linebuffer = (unsigned char*)malloc(width)) == NULL)
		return 0;
	// init decompressor

	while(1) {
		if(bitsleft == 8) {
			if(++p >= q && (((blocksize = fgetc(fp)) < 1) ||
				(q = (p = b) + fread(b, 1, blocksize, fp)) < (b + blocksize))) {
				free(linebuffer);
				return 0;
			}
			bitsleft = 0;
		}
		//
		thiscode = *p;
		//
		if((currentcode = (codesize + bitsleft)) <= 8) {
			*p >>= codesize;
			bitsleft = currentcode;
		} else {
			if(++ p >= q && (((blocksize = fgetc(fp)) < 1)||
				(q = (p = b) + fread(b, 1, blocksize,fp)) < (b + blocksize))) {
				free(linebuffer);
				return 0;
			}
			thiscode |= *p << (8 - bitsleft);
			if(currentcode <=16)
				*p >>= (bitsleft = currentcode - 8);
			else {
				if(++ p >= q &&  (((blocksize = fgetc(fp)) < 1) ||
					(q = (p = b) + fread(b, 1, blocksize, fp)) < (b + blocksize))) {
					free(linebuffer);
	       			return 0;
				}
			thiscode |= *p << (16 - bitsleft);
			*p >>= (bitsleft = currentcode - 16);
			}
		}
		//
		thiscode &= wordmasktable[codesize];
		currentcode = thiscode;
		//
		if(thiscode == (bits2 + 1))
			break;
		//
		if(thiscode >  nextcode) {
			free(linebuffer);
	 		return 0;
		}
		//
		if(thiscode == bits2) {
			nextcode = bits2+2;
			codesize2 = 1 << (codesize = (bits + 1));
			oldtoken = oldcode = NO_CODE;
			continue;
		}
		//
		u = firstcodestack;
		//
		if(thiscode == nextcode) {
			if(oldcode == NO_CODE) {
				free(linebuffer);
				return 0;
			}
			*u ++ = oldtoken;
			thiscode = oldcode;
		}
		//
		while(thiscode >= bits2) {
			*u ++ = lastcodestack[thiscode];
			thiscode = codestack[thiscode];
		}
		//
		oldtoken = thiscode;
		//
		while(1) {
			linebuffer[byte ++] = thiscode;
			//
			if(byte >= width) {
				for(xi = 0; xi < width; xi ++) {
					bytes[xi + Width * (line + Height - height)] = 
						(palette[linebuffer[xi] * 3 + 2] << 16) +
						(palette[linebuffer[xi] * 3 + 1] << 8) +
						palette[linebuffer[xi] * 3];
				}
				// here is image-copy loop
				byte = 0;
				//
				if(flags & 0x40) {
					line += inctable[pass];
					if(line >= height)
						line = startable[++ pass];
				} else
					++line;
			}
			if(u <= firstcodestack)
				break;
			//
			thiscode = *(-- u);
		}
		//
		if(nextcode < 4096 && oldcode != NO_CODE) {
			codestack[nextcode] = oldcode;
			lastcodestack[nextcode] = oldtoken;
			if(++ nextcode >= codesize2 && codesize < 12)
				codesize2 = 1 << ++ codesize;
		}
		//
		oldcode = currentcode;
	}

	return 1;
}

GIFHEADER Read_GIFHeader(FILE *f)
{
	GIFHEADER GIFHead;

	fread(&GIFHead.signature, 1, 3, f);
	fread(&GIFHead.version, 1, 3, f);
	fread(&GIFHead.screenwidth, 2, 1, f);
	fread(&GIFHead.screenheight, 2, 1, f);
	fread(&GIFHead.flags, 1, 1, f);
	fread(&GIFHead.background, 1, 1, f);
	fread(&GIFHead.aspect, 1, 1, f);

	return GIFHead;
}

IMAGEBLOCK Read_IBHeader(FILE *f)
{
	IMAGEBLOCK IBHead;

	fread(&IBHead.left, 2, 1, f);
	fread(&IBHead.top, 2, 1, f);
	fread(&IBHead.width, 2, 1, f);
	fread(&IBHead.height, 2, 1, f);
	fread(&IBHead.flags, 1, 1, f);

	return IBHead;
}

unsigned char *Read_GIFDataBlock(FILE *f)
{
    unsigned char *ptr1;
    unsigned char *ptr2;
    unsigned char dataSize;
    unsigned short bufSize;

    bufSize = 0;

    dataSize = fgetc(f);
	// Get the size of the first sub-block
    
    if((ptr1 = (unsigned char*) malloc(dataSize + 1)) == NULL)
        return NULL;
	// allocate initial data buffer
	ptr2 = ptr1;

    while(1) {
        bufSize += (dataSize + 1);
		// Running total of the buffer size

        *ptr1 ++ = dataSize;
		// Write the data count

        while(dataSize --)
            *ptr1 ++ = fgetc(f);
		/* Read/write the Plain Text data */
        
        if((dataSize = fgetc(f)) == 0)
            break;
		// Check if there is another data sub-block
		// (if dataSize = 0 -> Block Terminator encountered)

        if((ptr1 = (unsigned char*) realloc(ptr2, dataSize + 1)) == NULL)
            return NULL;
		// Increase the buffer size to accomodate the next sub-block
		ptr2 = ptr1;

        ptr1 += bufSize;
		// ptr1 points to the end of block
    }
    *ptr1++ = 0;
	// add null terminator

    return ptr2;
}

GIFPLAINTEXT ReadGifPlainText(FILE *f)
{
	GIFPLAINTEXT GIFPlainText;

	fread(&GIFPlainText.BlockSize, 1, 1, f);
	fread(&GIFPlainText.TextGridLeft, 2, 1, f);
	fread(&GIFPlainText.TextGridTop, 2, 1, f);
	fread(&GIFPlainText.TextGridWidth, 2, 1, f);
	fread(&GIFPlainText.TextGridHeight, 2, 1, f);
	fread(&GIFPlainText.CellWidth, 1, 1, f);
	fread(&GIFPlainText.CellHeight, 1, 1, f);
	fread(&GIFPlainText.TextFgColorIndex, 1, 1, f);
	fread(&GIFPlainText.TextBgColorIndex, 1, 1, f);
	//
	GIFPlainText.Terminator = 0;
	GIFPlainText.PlainTextData = Read_GIFDataBlock(f);

	return GIFPlainText;
}

GIFCOMMENT ReadGifComment(FILE *f)
{
	GIFCOMMENT GIFComment;

	GIFComment.CommentData = Read_GIFDataBlock(f);
	GIFComment.Terminator = 0;

	return GIFComment;
}

GIFGRAPHICCONTROL ReadGifGraphicControl(FILE *f)
{
	GIFGRAPHICCONTROL GIFGraphCtrl;

	fread(&GIFGraphCtrl.BlockSize, 1, 1, f);
	fread(&GIFGraphCtrl.PackedField, 1, 1, f);
	fread(&GIFGraphCtrl.DelayTime, 2, 1, f);
	fread(&GIFGraphCtrl.ColorIndex, 1, 1, f);
	fread(&GIFGraphCtrl.Terminator, 1, 1, f);

	return GIFGraphCtrl;
}

GIFAPPLICATION ReadGifApplication(FILE *f)
{
	GIFAPPLICATION GIFApplication;

	fread(&GIFApplication.BlockSize, 1, 1, f);
	fread(&GIFApplication.Identifier, 8, 1, f);
	fread(&GIFApplication.AuthentCode, 3, 1, f);
	//
	GIFApplication.ApplicationData = Read_GIFDataBlock(f);
	GIFApplication.Terminator = 0;

	return GIFApplication;
}

// nahraje gif, pokud je animovanej, ulozi frejmy pod sebe...
// zatim to fachalo pro vsecky gify co sem zkousel (bannery z netu,
// veci co mam na disku..)
BMP *Load_GIF(char *path)
{
	GIFGRAPHICCONTROL gifGraphicControl;
	GIFAPPLICATION gifApplication;
	GIFPLAINTEXT gifPlainText;
	GIFCOMMENT gifComment;
	int width, height, flags;
	int b, c, id, bits;
	IMAGEBLOCK iblk;
	GIFHEADER gh;
	FILE *GIFfp;
	BMP *tmp;

	if((GIFfp = fopen(path, "rb")) == NULL)
		return NULL;
	// open file

	Height = 0;
	bytes = (unsigned __int32*) malloc(0);

	gh = Read_GIFHeader(GIFfp);
	// read header

	if(memcmp(gh.signature, "GIF", 3)) {
		fclose(GIFfp);
		return NULL;
	}
	// check image is gif

	if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
		return NULL;
	// alloc mem for bitmap

	bits = (gh.flags & 0x07) + 1;
	// bpp
   
	if(gh.flags & 0x80) {
		c = 3 * (1 << ((gh.flags & 0x07) + 1));
		// size of palette (24bpp)

		fread(palette, 1, c, GIFfp);
		// read palette
	}

	while((c = fgetc(GIFfp)) == ',' || c == '!' || c == 0) {
		if(c == ',') { // 0x2c - image descriptor
			iblk = Read_IBHeader(GIFfp);
			//
			width = iblk.width;
			height = iblk.height;
			Width = width;
			Height += height;
			if((bytes = (unsigned __int32*) realloc(bytes, 4 * Width * Height)) == NULL) {
				fclose(GIFfp);
				return NULL;
			}
			// update bitmap data

			if(iblk.flags & 0x80) {
				b = 3 * (1 << ((iblk.flags & 0x07) + 1));
				fread(palette, 1, b, GIFfp);
				bits = (iblk.flags & 0x07) + 1;
			}
			// read private color table if present

			if((c = fgetc(GIFfp)) == EOF) {
				fclose(GIFfp);
				return NULL;
			}
			// read block marker

			flags = iblk.flags;
			UnpackImage(GIFfp, c, width, height, flags);
			// decompress gif frame
		} else if(c == '!') { // 0x21 - extension block
			id = fgetc(GIFfp);
			//
            switch(id) {
            case 0x01: // Plain Text Extension
                gifPlainText = ReadGifPlainText(GIFfp);
                break;
            case 0xfe: // Comment Extension
                gifComment = ReadGifComment(GIFfp);
                break;
            case 0xf9: // Graphic Control Extension
                gifGraphicControl = ReadGifGraphicControl(GIFfp);
                break;
			case 0xff: // 0xff - app extension
				gifApplication = ReadGifApplication(GIFfp);
				break;
			}
		} else if(c == ';') { // 0x3b - trailer
			fclose(GIFfp);
			return NULL;
		}
	}

	fclose(GIFfp);
	// close file

	tmp->width = Width;
	tmp->height = Height;
	tmp->bytes = bytes;
	// update bitmap data

	if(Height == 0) {
		free((void*)tmp->bytes);
		free((void*)tmp);
		//
		return NULL;
	}

	return tmp;
}


