/*
						+--------------------------------+
                        |                                |
                        |      *** MipMap Maker ***      |
                        |                                |
                        |          L.Polok 2002         |
                        |                                |
                        |             Main.c             |
                        |                                |
						+--------------------------------+
*/

#include <string.h>
#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
#include <math.h>
#include "True_Bmp.h"
#include "True_Pcx.h"
#include "256_Bmp.h"
#include "256_Pcx.h"
#include "Gif.h"
#include "Tga.h"
#include "FileFormat.h"

// nahraje obrazek podporovanych formatu do struktury BMP
// (jen pole bajtu + vyska + sirka rastru > True_BMP.h)
BMP *LoadThePicture(char *str)
{
	BMP *tex;

	switch(DetectFormat(str)) {
	case (FFORMAT_BMP):
		tex = Load_True_ColorBMP(str);
		if(tex == NULL)
			tex = Load_256_ColorBMP(str);
		break;
	case (FFORMAT_PCX):			
		tex = Load_True_ColorPCX(str);
		if(tex == NULL)
			tex = Load_256_ColorPCX(str);
		break;
	case (FFORMAT_GIF):			
		tex = Load_GIF(str); // pokud je anim. ulozi frejmy nad sebe...
		break;
	case (FFORMAT_TGA):			
		tex = Load_TGA(str);
		break;
	case (FFORMAT_UNKNOWN):
	default: 
		tex = NULL;
		break;
	}

	return tex;
}

// ulozi rastr jako truecolorovou bitmapu
// (nevim do jake miry to facha, ale precte to
// paintbrush, acd-see i Load_TrueColorBMP() ...)
int	Save_TrueColor_BMP(char *filename, BMP *bmp)
{
	unsigned long ByteCount;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	unsigned char *lpbBuf;
	unsigned __int32 bBuf;
	unsigned int BuffLen;
	int i, j, x;
	FILE *BMPfp;

	BuffLen = ((3 * bmp->width + 3) >> 2) << 2;
	lpbBuf =  (unsigned char*) malloc(BuffLen);
	// compute lenght of line-buffer

	if((BMPfp = fopen(filename, "wb")) == NULL)
		return 0;
	// opens file

	bmfh.bfType = ('B'|('M'<<8)); // "BM"
	bmfh.bfSize = 0;              // necessary to modify
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER);
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = bmp->width;
	bmih.biHeight = bmp->height;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;	      // True Color
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = 0;         // necessary to modify
	bmih.biXPelsPerMeter = 96;
	bmih.biYPelsPerMeter = 96;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	// create header

	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);
	// writing header

	ByteCount = 0;
	for(i = bmp->height - 1; i >= 0; i --) {
		for(j = 0, x = 0; j < bmp->width; j ++, x ++) {
			bBuf = bmp->bytes[i * bmp->width + j];
			lpbBuf[x * 3] = (char)((bBuf >> 16) & 0xff);
			lpbBuf[x * 3 + 1] = (char)((bBuf >> 8) & 0xff);
			lpbBuf[x * 3 + 2] = (char)(bBuf & 0xff);
			// saved in BGR, not RGB
			ByteCount += 3;
		}

		if(fwrite(lpbBuf, 1, BuffLen, BMPfp) != BuffLen) {
			free(lpbBuf);
			return 0;
		}
	}

	// repair header
	bmih.biSizeImage = ByteCount;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) +
	   ByteCount + sizeof(BITMAPINFOHEADER);
	fseek(BMPfp, 0, SEEK_SET);
	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);

	fclose(BMPfp);

	free(lpbBuf);

	return 1;
}

// vytiskne logo programu
void Logo()
{
	//printf("         1         2         3         4         5         6         7         8");
	//printf("12345678901234567890123456789012345678901234567890123456789012345678901234567890");
	printf("                          _______MipMap maker_______                         \n"
		   "                        L.Polok (c) 2002                       \n"
		   "                   Reads .pcx, .bmp, .gif, .tga writes .bmp                  \n"
		   "                see www.luki.webzdarma.cz / luki.sach@atlas.cz           \n\n\n");
	// vypise logo (je uprostred za predpokladu ze radek ma 80 znaku :-) )
}

// vytvori jmeno mipmapy
void Create_Filename(char *src, char *dest)
{
	char buffer[256];
    char *pt;

	strcpy(buffer, src);
	// buffer = "test.bmp"
	pt = strrchr(buffer, '.');
	if(pt != NULL)
		*pt = 0;
	// buffer = "test"
	strcat(buffer, "_mip.bmp");
	// buffer = "test_mip.bmp"

	strcpy(dest, buffer);
}

// vyextrakti z adresy jen meno souboru
void Short_Filename(char *src, char *dest)
{
    char *pt;

	// src = "c:\somefolder\test.bmp"
	pt = strrchr(src, '\\');
	if(pt != NULL) {
		// pt = "\test.bmp"
		pt ++;
		strcpy(dest, pt);
		//
		return;
	}

	strcpy(dest, src);
}

// zmensi bitmapu po x-ove ose
int Stretch_X(BMP *bitmap, int x, int y, int w, int h)
{
	unsigned __int32 a, b, c;
	int i, j;

	if(x + w + w / 2 > bitmap->width || y + h > bitmap->height)
		return 0;
	// check size, pos

	for(j = 0; j < h; j ++) {
		for(i = 0; i < w; i += 2) {
			a = bitmap->bytes[(x + i) + (y + j) * bitmap->width];
			b = bitmap->bytes[(x + i + 1) + (y + j) * bitmap->width];
			// get pixels
			c = ((((a & 0xff00ff) + (b & 0xff00ff)) >> 1) & 0xff00ff) +
				((((a & 0x00ff00) + (b & 0x00ff00)) >> 1) & 0x00ff00);
			// mix color
			bitmap->bytes[(x + w + i / 2) + (y + j) * bitmap->width] = c;
		}
	}
	// stretch

	return 1;
}

// zmensi bitmapu po y-ove ose
int Stretch_Y(BMP *bitmap, int x, int y, int w, int h)
{
	unsigned __int32 a, b, c;
	int i, j;

	if(x + w > bitmap->width || y + h + h / 2 > bitmap->height)
		return 0;
	// check size, pos

	for(j = 0; j < h; j += 2) {
		for(i = 0; i < w; i ++) {
			a = bitmap->bytes[(x + i) + (y + j) * bitmap->width];
			b = bitmap->bytes[(x + i) + (y + 1 + j) * bitmap->width];
			// get pixels
			c = ((((a & 0xff00ff) + (b & 0xff00ff)) >> 1) & 0xff00ff) +
				((((a & 0x00ff00) + (b & 0x00ff00)) >> 1) & 0x00ff00);
			// mix color
			bitmap->bytes[(x + i) + (y + h + j / 2) * bitmap->width] = c;
		}
	}
	// stretch

	return 1;
}

// zmensi bitmapu po obou stranach
int Stretch_Both(BMP *bitmap, int x, int y, int w, int h)
{
	unsigned __int32 a, b, c, d, mix;
	int i, j;

	if(x + w + w / 2 > bitmap->width || y + h > bitmap->height)
		return 0;
	// check size, pos

	for(j = 0; j < h; j += 2) {
		for(i = 0; i < w; i += 2) {
			a = bitmap->bytes[(x + i) + (y + j) * bitmap->width];
			b = bitmap->bytes[(x + i + 1) + (y + j) * bitmap->width];
			c = bitmap->bytes[(x + i) + (y + 1 + j) * bitmap->width];
			d = bitmap->bytes[(x + i + 1) + (y + 1 + j) * bitmap->width];
			// get pixels
			mix = ((((a & 0xff00ff) + (b & 0xff00ff) +
				(c & 0xff00ff) + (d & 0xff00ff)) >> 2) & 0xff00ff) +
				((((a & 0x00ff00) + (b & 0x00ff00) +
				(c & 0x00ff00) + (d & 0x00ff00)) >> 2) & 0x00ff00);
			// mix color
			bitmap->bytes[(x + w + i / 2) + (y + j / 2) * bitmap->width] = mix;
		}
	}
	// stretch

	return 1;
}

void CreateBiMipMap(BMP *bitmap)
{
	int x, y, w, h;
	int lw, lh;
	int lx, ly;

	if(bitmap->width <= 1 || bitmap->height <= 1)
		return;
	// kontroluje velikost obrazku

	lw = bitmap->width / 2;
	lh = bitmap->height / 2;
	lx = 0;
	ly = 0;

	while(lw != 0 && lh != 0) {
		x = lx;
		y = ly;
		w = lw;
		h = lh;
		while(w) {
			if(!Stretch_X(bitmap, x, y, w, h))
				break;
			x += w;
			w /= 2;
		}
		//
		x = lx;
		y = ly;
		w = lw;
		h = lh;
		while(h) {
			if(!Stretch_Y(bitmap, x, y, w, h))
				break;
			y += h;
			h /= 2;
		}
		//
		x = lx;
		y = ly;
		w = lw;
		h = lh;
		Stretch_Y(bitmap, x + w, y, w / 2, h);
		//
		lx += lw;
		ly += lh;
		lw /= 2;
		lh /= 2;
	}
	// vytvori bi-directional mipmap
}

void Create_MipMap(BMP *bitmap)
{
	int x, y, w, h;

	if(bitmap->width <= 1 || bitmap->height <= 1)
		return;
	// kontroluje velikost obrazku

	x = 0;
	y = 0;
	w = bitmap->width / 2;
	h = bitmap->height;

	while(w || h) {
		Stretch_Both(bitmap, x, y, w, h);
		x += w;
		w /= 2;
		h /= 2;
	}
	// vytvori ordinary mipmapu..
}

void main(int argc, char **argv)
{
	BMP *texture, mipped;
	char outfile[256];
	int i;

	Logo();

	if(argc < 2) {
		printf("Program must be executed with commandline :\n"
			   "    Mip_Maker.exe [filename] <ord>\n\n"
			   "Where filename is name of texture (necessary)\n"
			   "And if there will be next parameter, program generates\n"
			   "ordinary mip-map only.\n\n"
			   "Output is in truecolor bitmap with name :\n"
			   "    \"Original_Filename\" + \"_mip\" + \".bmp\"\n\n"
			   "Press any key, Allah sees ya ! <|:-)=\n");
		//
		getchar();
		return;
	}

	Short_Filename(argv[1], &(outfile[0]));
	printf("Loading image \'%s\' ...", outfile);
	if((texture = LoadThePicture(argv[1])) == NULL) {
		printf("Invalid fileformat !\n"
			   "Press any key, Allah sees ya ! <|:-)=\n");
		//
		getchar();
		return;
	}
	printf(" Done\n");
	// tady se nahraje obrazek

	// ...........
	printf("Creating %s MipMap ...", (argc > 2)? "ordinary" : "bi-directional");
	//
	mipped.width = texture->width * 2;
	mipped.height = texture->height * ((argc > 2)? 1 : 2);
	if((mipped.bytes = (unsigned __int32*) malloc
	   (mipped.width * mipped.height * 4)) == NULL) {
		printf("Not enough memory for MipMap !\n");
		getchar();
		return;
	}
	// vytvori se bitmapa pro mip-mapu

	memset(mipped.bytes, 0, mipped.width * mipped.height * 4);
	for(i = 0; i < texture->height; i ++) {
		memcpy(&mipped.bytes[i * mipped.width],
			&texture->bytes[i * texture->width], texture->width * 4);
	}
	// bitmapa se zkopiruje do levyho horniho rohu mip-mapy

	if(argc == 2)
		CreateBiMipMap(&mipped);
	else
		Create_MipMap(&mipped);
	// tady se dela mip-mapa ...
	printf(" Done\n");

	Create_Filename(argv[1], &(outfile[0]));
	printf("Writing Mip-Map ...");
	Save_TrueColor_BMP(outfile, &mipped);
	printf(" Done\n\n\n");
	// ulozi se na disk

	printf("Press any key, Allah sees ya ! <|:-)=\n");
	getchar();
}